/*

Copyright 2022 Piers Harding
Copyright 2022 SKA Observatory

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

*/

package models

import (
	"encoding/json"

	"github.com/appscode/go/log"
	dtypes "gitlab.com/piersharding/tango-operator/types"
	"gitlab.com/piersharding/tango-operator/utils"
	batchv1 "k8s.io/api/batch/v1"
)

// DeviceServerConfiguratorJob generates the Job description for
// the DeviceServer Job
func DeviceServerConfiguratorJob(dcontext dtypes.OperatorContext) (*batchv1.Job, error) {

	const deviceServerJob = `
apiVersion: batch/v1
kind: Job
metadata:
  name: deviceserverconfigurator-job-{{ .Name }}
  namespace: {{ .Namespace }}
  labels:
    app.kubernetes.io/name: deviceserverconfigurator-job
    app.kubernetes.io/instance: "{{ .Name }}"
    app.kubernetes.io/managed-by: DeviceServerController
spec:
  ttlSecondsAfterFinished: 100
  backoffLimit: 2
  completions: 1
  parallelism: 1
  template:
    metadata:
      labels:
        app.kubernetes.io/name: deviceserverconfigurator-job
        app.kubernetes.io/instance: "{{ .Name }}"
        app.kubernetes.io/managed-by: DeviceServerController
    spec:
      serviceAccountName: "deviceserverconfigurator-serviceaccount-{{ .Name }}"
      restartPolicy: Never
    {{- with .PullSecrets }}
      imagePullSecrets:
      {{range $val := .}}
      - name: {{ $val.name }}
      {{end}}
	  {{- end }}


      containers:
      - name: configurator
		securityContext:
		  runAsUser: 0
		image: "{{ .DSConfigImage }}"
        imagePullPolicy: {{ .PullPolicy }}
        command: # exit code 2 is CONFIG_APPLIED - https://github.com/MaxIV-KitsControls/lib-maxiv-dsconfig/blob/master/dsconfig/utils.py#L11 !!!! this should not be an error !!!!
          - sh
          - data/dsconfig.sh
        env:
        - name: TANGO_HOST
          value: {{ .DatabaseDS }}:{{ .Port }}
                {{- with .Env }}
  {{ toYaml . | indent 8 }}
  {{- end }}
        volumeMounts:
          - name: configuration
            mountPath: data
            readOnly: true
      volumes:
        - name: configuration
          configMap:
            defaultMode: 0777
            name: "deviceserver-configs-{{ .Name }}"
{{- with .Volumes }}
{{ toYaml . | indent 6 }}
{{- end }}
{{- with .NodeSelector }}
      nodeSelector:
{{ toYaml . | indent 8 }}
{{- end }}
{{- with .Affinity }}
      affinity:
{{ toYaml . | indent 8 }}
{{- end }}
{{- with .Tolerations }}
      tolerations:
{{ toYaml . | indent 8 }}
{{- end }}
      restartPolicy: Never
`

	result, err := utils.ApplyTemplate(deviceServerJob, dcontext)
	if err != nil {
		log.Debugf("ApplyTemplate Error: %+v\n", err)
		return nil, err
	}

	job := &batchv1.Job{}
	if err := json.Unmarshal([]byte(result), job); err != nil {
		return nil, err
	}
	return job, err
}
